# valid addresses
P2PKH_ADDRESS = "1H36XJJRdzbTJ8yqcR6HBvSTuJJ5dvf5wC"
P2SH_ADDRESS = "3CKnc9a1da7Nnh3SEzvVZe7LfkBq2pYTM2"
BECH32_ADDRESS = "bc1qpfa829hg70yhjw7sg8t95vjtz7gg38k3kfm3ny"

# valid testnet addresses
P2PKH_ADDRESS_TESTNET = "mwZ3pMPQT22i5FTTKz4f1qenmHtncVQqCV"
P2SH_ADDRESS_TESTNET = "2N3szftW3F2cizUfyv8YNBb6bt6PzsUTi92"
BECH32_ADDRESS_TESTNET = "tb1qdh780ud5vtn00327z7qzhjd9nv5fm8qu6qkg2d"

# invalid addresses
INVALID_P2PKH_ADDRESS = "1BvBMSEYstWetqTFn5Au4m4GF07xJaNVN2"
INVALID_P2SH_ADDRESS = "3CKnc9a1da7Nnh3SEzvVZe7LfkBq2IYTM2"
INVALID_BECH32_ADDRESS = "bc1qpfa829hg70yhjw7sg8t95vjtz7gg38k3kf03ny"

# mnemonic phrases
MNEMONIC_1 = "behind view duty stairs globe provide tired poem balance advance early husband unhappy final crunch parade cruise jacket trap bread mention cannon decide alpha"
MNEMONIC_2 = "solid favorite loop sustain grit hub attitude empower prize push extend humor"
MNEMONIC_3 = "chalk range lounge edit arrive erupt dirt voyage buffalo practice find ecology"
